"""Main module of Lexicon. Defines the Client class, that holds all Lexicon logic."""
from __future__ import absolute_import
import importlib

import tldextract

from lexicon import discovery
from lexicon.config import (
    ConfigResolver, DictConfigSource,
    legacy_config_resolver, non_interactive_config_resolver,
)


class ProviderNotAvailableError(Exception):
    """
    Custom exception to raise when a provider is not available,
    typically because some optional dependencies are missing
    """


class Client(object):  # pylint: disable=useless-object-inheritance,too-few-public-methods
    """This is the Lexicon client, that will execute all the logic."""

    def __init__(self, config=None):
        if not config:
            # If there is not config specified, we load a non-interactive configuration.
            self.config = non_interactive_config_resolver()
        elif not isinstance(config, ConfigResolver):
            # If config is not a ConfigResolver, we are in a legacy situation.
            # We protect this part of the Client API.
            self.config = legacy_config_resolver(config)
        else:
            self.config = config

        # Validate configuration
        self._validate_config()

        runtime_config = {}

        # Process domain, strip subdomain
        domain_parts = tldextract.extract(
            self.config.resolve('lexicon:domain'))
        runtime_config['domain'] = '{0}.{1}'.format(
            domain_parts.domain, domain_parts.suffix)

        if self.config.resolve('lexicon:delegated'):
            # handle delegated domain
            delegated = self.config.resolve('lexicon:delegated').rstrip('.')
            if delegated != runtime_config.get('domain'):
                # convert to relative name
                if delegated.endswith(runtime_config.get('domain')):
                    delegated = delegated[:-len(runtime_config.get('domain'))]
                    delegated = delegated.rstrip('.')
                # update domain
                runtime_config['domain'] = '{0}.{1}'.format(
                    delegated, runtime_config.get('domain'))

        self.action = self.config.resolve('lexicon:action')
        self.provider_name = (self.config.resolve('lexicon:provider_name')
                              or self.config.resolve('lexicon:provider'))

        self.config.add_config_source(DictConfigSource(runtime_config), 0)

        provider_module = importlib.import_module(
            'lexicon.providers.' + self.provider_name)
        provider_class = getattr(provider_module, 'Provider')
        self.provider = provider_class(self.config)

    def execute(self):
        """Execute provided configuration in class constructor to the DNS records"""
        self.provider.authenticate()
        identifier = self.config.resolve('lexicon:identifier')
        record_type = self.config.resolve('lexicon:type')
        name = self.config.resolve('lexicon:name')
        content = self.config.resolve('lexicon:content')

        if self.action == 'create':
            return self.provider.create_record(record_type, name, content)

        if self.action == 'list':
            return self.provider.list_records(record_type, name, content)

        if self.action == 'update':
            return self.provider.update_record(identifier, record_type, name, content)

        if self.action == 'delete':
            return self.provider.delete_record(identifier, record_type, name, content)

        raise ValueError('Invalid action statement: {0}'.format(self.action))

    def _validate_config(self):
        provider_name = self.config.resolve('lexicon:provider_name')
        if not self.config.resolve('lexicon:provider_name'):
            raise AttributeError('provider_name')

        try:
            available = discovery.find_providers()[self.config.resolve('lexicon:provider_name')]
        except KeyError:
            raise ProviderNotAvailableError('This provider ({0}) is not supported by Lexicon.'
                                            .format(provider_name))
        else:
            if not available:
                raise ProviderNotAvailableError(
                    'This provider ({0}) has required dependencies that are missing. '
                    'Please install lexicon[{0}] first.'.format(provider_name))

        if not self.config.resolve('lexicon:action'):
            raise AttributeError('action')
        if not self.config.resolve('lexicon:domain'):
            raise AttributeError('domain')
        if not self.config.resolve('lexicon:type'):
            raise AttributeError('type')



"""Integration tests for auto"""
import socket
from unittest import TestCase

import mock
import pytest
from lexicon.tests.providers.integration_tests import IntegrationTests
from lexicon.providers.auto import _get_ns_records_domains_for_domain


# This fixture ensures to mock _get_ns_records_domains_for_domain, in order to not rely
# on the machine on which the test is done, as this function call nslookup.
# Then it will prevent errors where there is no network or tested domain do not exists anymore.
@pytest.fixture(autouse=True)
def _nslookup_mock(request):
    _ignore_nslookup_mock = request.node.get_closest_marker('ignore_nslookup_mock')

    if _ignore_nslookup_mock:
        # Do not mock if the test says so.
        yield
    else:
        with mock.patch('lexicon.providers.auto._get_ns_records_for_domain',
                        return_value=['ns.ovh.net']) as fixture:
            yield fixture


# Guys, are we online ?
def _there_is_no_network():
    try:
        socket.create_connection(("www.google.com", 80))
        return False
    except (OSError, IOError):
        pass
    return True


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class AutoProviderTests(TestCase, IntegrationTests):
    """TestCase for auto"""
    provider_name = 'auto'
    domain = 'pacalis.net'

    def _filter_headers(self):
        return ['X-Ovh-Application', 'X-Ovh-Consumer', 'X-Ovh-Signature']

    def _test_parameters_overrides(self):
        return {'auth_entrypoint': 'ovh-eu'}

    def _test_fallback_fn(self):
        return lambda x: 'placeholder_' + x if x != 'mapping_override' else None

    # Here we do not mock the function _get_ns_records_domains_for_domain
    # to effectively test the nslookup call and processing.\
    @pytest.mark.skipif(_there_is_no_network(), reason='No network, no nslookup call possible.')
    @pytest.mark.ignore_nslookup_mock('yes')
    def test_nslookup_resolution(self):  #  pylint: disable=no-self-use
        """Ensure that nameservers can be resolved through os nslookup call."""
        assert _get_ns_records_domains_for_domain('google.com')



"""Integration tests for Onapp"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTests


class OnappProviderTests(TestCase, IntegrationTests):
    """TestCase for Oneapp"""
    provider_name = 'onapp'
    domain = 'my-test.org'

    def _filter_headers(self):
        return ['Authorization']

    def _test_parameters_overrides(self):
        return {'auth_server': 'https://dashboard.dynomesh.com.au'}



"""Module provider for nfsn"""
from __future__ import absolute_import, print_function
import hashlib
import logging
import random
import string
import time
try:
    from urllib.parse import urlencode
except ImportError:
    from urllib import urlencode

import requests
from lexicon.providers.base import Provider as BaseProvider


LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['nearlyfreespeech.net']


def provider_parser(subparser):
    """Generate subparser for nfsn"""
    subparser.add_argument(
        "--auth-username", help="specify username used to authenticate")
    subparser.add_argument(
        "--auth-token", help="specify token used to authenticate")


SALT_SHAKER = string.ascii_letters + string.digits


class Provider(BaseProvider):
    """Provider class for nfsn"""
    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = 'https://api.nearlyfreespeech.net'
        self.shortname = None

    def _authenticate(self):
        self._post('/dns/{0}/listRRs'.format(self.domain))
        self.domain_id = self.domain

    # Create record. If record already exists with the same content, do nothing'
    def _create_record(self, rtype, name, content):
        existing_records = self.list_records(rtype, name, content)
        if existing_records:
            # Do nothing if the record already exists.
            # The creation call can fail for a variety of reasons, so
            # the safest thing to do is check if the record already exists
            return True

        self._do_create(rtype, name, content)
        LOGGER.debug('create_record: %s', True)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        params = {}
        if rtype is not None:
            params['type'] = rtype
        if name is not None:
            params['name'] = self._relative_name(name)
        if content is not None:
            params['data'] = content

        url = '/dns/{0}/listRRs'.format(self.domain_id)
        records = self._post(url, params)
        records = [{
            'type': r['type'],
            'name': self._full_name(r['name']),
            'ttl': r['ttl'],
            'content': r['data'],
            'id': hashlib.sha1(''.join(
                [r['type'], r['name'], r['data']]).encode('utf-8')).hexdigest()
        } for r in records]

        LOGGER.debug('list_records: %s', records)
        return records

    # Create or update a record.

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        if identifier is not None:
            records = self._list_records()
            to_delete = next(
                (r for r in records if r['id'] == identifier), None)
            if to_delete is None:
                raise ValueError('No record with that identifier.')
        else:
            # Check name and rtype
            matching_records = self._list_records(rtype=rtype, name=name)
            if len(matching_records) > 1:
                raise ValueError(
                    'More than one record exists with that type and name. '
                    'Try specifying an identifier.')
            to_delete = matching_records[0]

        self._do_delete(to_delete['type'],
                        to_delete['name'], to_delete['content'])
        self._do_create(rtype, name, content)
        LOGGER.debug('update_record: %s', True)
        return True

    # Delete an existing record
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        matching_records = self._list_records(rtype, name, content)
        if identifier is not None:
            to_delete = next(
                (r for r in matching_records if r['id'] == identifier), None)
            if to_delete is None:
                raise ValueError('No record with that identifier.')
            to_delete = [to_delete]
        else:
            to_delete = matching_records

        for record in to_delete:
            self._do_delete(record['type'], record['name'], record['content'])

        LOGGER.debug('delete_record: %s', True)
        return True

    def _do_create(self, rtype, name, content):
        record = {
            'name': self._relative_name(name),
            'type': rtype,
            'data': content
        }

        ttl = self._get_lexicon_option('ttl')
        if ttl:
            record['ttl'] = ttl

        self._post('/dns/{0}/addRR'.format(self.domain_id), record)
        return True

    def _do_delete(self, rtype=None, name=None, content=None):
        url = '/dns/{0}/removeRR'.format(self.domain_id)
        record = {
            'name': self._relative_name(name),
            'type': rtype,
            'data': content
        }
        self._post(url, record)
        return True

    # Helpers
    def _request(self, action='GET', url='/', data=None, query_params=None):
        if data is not None:
            body = urlencode(data)
            hashed_body = hashlib.sha1(body.encode('utf-8')).hexdigest()
        else:
            hashed_body = hashlib.sha1().hexdigest()

        timestamp = str(int(time.time()))
        salt = ''.join(random.choice(SALT_SHAKER) for _ in range(16))
        hash_items = [
            self._get_provider_option('auth_username'),
            timestamp,
            salt,
            self._get_provider_option('auth_token'),
            url,
            hashed_body]
        auth_hash = hashlib.sha1(
            ';'.join(hash_items).encode('utf-8')).hexdigest()
        auth_value = ';'.join([self._get_provider_option(
            'auth_username'), timestamp, salt, auth_hash])
        auth_header = {
            'X-NFSN-Authentication': auth_value
        }

        response = requests.request(action, ''.join([self.api_endpoint, url]),
                                    data=data,
                                    headers=auth_header)
        response.raise_for_status()
        if response.content:
            return response.json()
        return {}



"""Integration tests for Dreamhost"""
from unittest import TestCase

import pytest
from lexicon.tests.providers.integration_tests import IntegrationTests
from lexicon.providers.dreamhost import Provider


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class DreamhostProviderTests(TestCase, IntegrationTests):
    """TestCase for Dreamhost"""
    provider_name = 'dreamhost'
    domain = 'lexicon-example.com'

    def _filter_query_parameters(self):
        return ['key']

    @pytest.mark.skip(reason="can not set ttl when creating/updating records")
    def test_provider_when_calling_list_records_after_setting_ttl(self):
        return

    def test_identifier(self):
        """Test _identifier"""
        # pylint: disable=protected-access
        # pylint: disable=no-self-use
        dreamhost_record = {'type': 'A', 'record': 'www.example.com', 'value': '1.2.3.4'}
        identifier = Provider._identifier(dreamhost_record)

        dreamhost_record_from_id = Provider._id_to_dreamhost_record(identifier)

        assert dreamhost_record_from_id == dreamhost_record

    def test_id_to_record(self):
        """Test _id_to_record and _record_to_dreamhost_record"""
        # pylint: disable=protected-access
        # pylint: disable=no-self-use
        dreamhost_record = {'type': 'A', 'record': 'www.example.com', 'value': '1.2.3.4'}

        identifier = Provider._identifier(dreamhost_record)

        record = Provider._id_to_record(identifier)
        dreamhost_record_from_id = Provider._record_to_dreamhost_record(record)

        assert dreamhost_record_from_id == dreamhost_record

    def test_id_to_dreamhost_record(self):
        """Test _id_to_dreamhost_record"""
        # pylint: disable=protected-access
        # pylint: disable=no-self-use
        dreamhost_record = {'type': 'A', 'record': 'www.example.com', 'value': '1.2.3.4'}

        identifier = Provider._identifier(dreamhost_record)

        dreamhost_record_from_id = Provider._id_to_dreamhost_record(identifier)

        assert dreamhost_record_from_id == dreamhost_record



"""Integration tests for Hover"""
import json
import re
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTests


class HoverProviderTests(TestCase, IntegrationTests):
    """TestCase for Hover"""
    provider_name = 'hover'
    domain = 'novuslex.com'
    domain_id = 'dom1127777'
    hoverauth = '0123456789abcdef0123456789abcdef'
    hover_session = '0123456789abcdef0123456789abcdef'

    def _filter_post_data_parameters(self):
        return ['username', 'password']

    def _filter_headers(self):
        return ['Cookie']

    def _filter_query_parameters(self):
        return ['hover_session', 'hoverauth']

    def _replace_auth(self, cookie):
        cookie = re.sub('hover_session=.*;',
                        'hover_session={};'.format(self.hover_session),
                        cookie)
        cookie = re.sub('hoverauth=.*;',
                        'hoverauth={};'.format(self.hoverauth),
                        cookie)
        return cookie

    def _filter_response(self, response):
        if 'basestring' not in globals():
            basestring = str

        if 'set-cookie' in response['headers']:
            if isinstance(response['headers']['set-cookie'], basestring):
                response['headers']['set-cookie'] = \
                    self._replace_auth(response['headers']['set-cookie'])
            else:
                for i, cookie in enumerate(response['headers']['set-cookie']):
                    response['headers']['set-cookie'][i] = self._replace_auth(cookie)

        try:
            filtered_body = json.loads(response['body']['string'].decode('UTF-8'))
        except ValueError:
            # Body is not json during authentication, so we let it through.
            # Helper function _request in hover.py will raise exception when
            # response is not json and it should be.
            pass
        else:
            # filter out my personal contact information
            if 'contact' in filtered_body:
                del filtered_body['contact']

            # if the response is listing all my domains then return a mock response
            if 'domains' in filtered_body and len(filtered_body['domains']) > 1:
                filtered_body['domains'] = [
                    {
                        'id': self.domain_id,
                        'domain_name': self.domain,
                        'status': 'active',
                    },
                ]

            response['body']['string'] = json.dumps(filtered_body).encode('UTF-8')
        return response
